#pylint: disable=invalid-name
#pylint: disable=too-many-public-methods
#pylint: disable=too-many-branches
#pylint: disable=no-member
#pylint: disable=protected-access
#pylint: disable=bare-except
"""
TMC_2209 stepper driver communication module
"""

import time
from enum import Enum
import math
import threading
try:
    from RPi import GPIO
except:
    from Mock import GPIO
from ._TMC_2209_logger import Loglevel



class Direction(Enum):
    """movement direction of the motor"""
    CCW = 0
    CW = 1


class MovementAbsRel(Enum):
    """movement absolute or relative"""
    ABSOLUTE = 0
    RELATIVE = 1


class MovementPhase(Enum):
    """movement phase"""
    STANDSTILL = 0
    ACCELERATING = 1
    MAXSPEED = 2
    DECELERATING = 3


class StopMode(Enum):
    """stopmode"""
    NO = 0
    SOFTSTOP = 1
    HARDSTOP = 2


def set_movement_abs_rel(self, movement_abs_rel):
    """
    set whether the movment should be relative or absolute by default.
    See the Enum MovementAbsoluteRelative

        Paramters:
            movement_abs_rel (enum): whether the movment should be relative or absolute
    """
    self._movement_abs_rel = movement_abs_rel



def get_current_position(self):
    """
    returns the current motor position in microsteps

        Returns:
            current_pos (bool): current motor position
    """
    return self._current_pos



def set_current_position(self, new_pos):
    """
    overwrites the current motor position in microsteps

        Parameters:
            new_pos (bool): new position
    """
    self._current_pos = new_pos



def set_max_speed(self, speed):
    """
    sets the maximum motor speed in µsteps per second

        Parameters:
            speed (int): maximum speed in microsteps/sec
    """
    if speed < 0.0:
        speed = -speed
    if self._max_speed != speed:
        self._max_speed = speed
        self._cmin = 1000000.0 / speed
        # Recompute _n from current speed and adjust speed if accelerating or cruising
        if self._n > 0:
            self._n = (self._speed * self._speed) / (2.0 * self._acceleration) # Equation 16
            self.compute_new_speed()



def set_max_speed_fullstep(self, speed):
    """
    sets the maximum motor speed in fullsteps per second

        Parameters:
            speed (int): maximum speed in fullsteps/sec
    """
    self.set_max_speed(speed*self.get_microstepping_resolution())



def get_max_speed(self):
    """
    returns the maximum motor speed in steps per second

        Returns:
            max_speed (int): current maximum speed in steps/sec
    """
    return self._max_speed



def set_acceleration(self, acceleration):
    """
    sets the motor acceleration/decceleration in µsteps per sec per sec

        Parameters:
            acceleration (int): acceleration/decceleration in µsteps per sec per sec
    """
    if acceleration == 0.0:
        return
    acceleration = abs(acceleration)
    if self._acceleration != acceleration:
        # Recompute _n per Equation 17
        self._n = self._n * (self._acceleration / acceleration)
        # New c0 per Equation 7, with correction per Equation 15
        self._c0 = 0.676 * math.sqrt(2.0 / acceleration) * 1000000.0 # Equation 15
        self._acceleration = acceleration
        self.compute_new_speed()



def set_acceleration_fullstep(self, acceleration):
    """
    sets the motor acceleration/decceleration in fullsteps per sec per sec

        Parameters:
            acceleration (int): acceleration/decceleration in fullsteps per sec per sec
    """
    self.set_acceleration(acceleration*self.get_microstepping_resolution())



def get_acceleration(self):
    """
    returns the motor acceleration/decceleration in steps per sec per sec

        Returns:
            acceleration (int): acceleration/decceleration in µsteps per sec per sec
    """
    return self._acceleration



def stop(self, stop_mode = StopMode.HARDSTOP):
    """
    stop the current movement

        Parameters:
            stop_mode (enum): whether the movement should be stopped immediatly or softly
    """
    self._stop = stop_mode



def get_movement_phase(self):
    """
    return the current Movement Phase

        Returns:
            movement_phase (enum): current Movement Phase
    """
    return self._movement_phase



def run_to_position_steps(self, steps, movement_abs_rel = None):
    """
    runs the motor to the given position.
    with acceleration and deceleration
    blocks the code until finished or stopped from a different thread!
    returns true when the movement if finshed normally and false,
    when the movement was stopped

        Parameters:
            steps (int): amount of steps; can be negative
            movement_abs_rel (enum): whether the movement should be absolut or relative

        Returns:
            stop (enum): how the movement was finished
    """
    if movement_abs_rel is None:
        movement_abs_rel = self._movement_abs_rel

    if movement_abs_rel == MovementAbsRel.RELATIVE:
        self._target_pos = self._current_pos + steps
    else:
        self._target_pos = steps

    self._stop = StopMode.NO
    self._step_interval = 0
    self._speed = 0.0
    self._n = 0
    self.compute_new_speed()
    while self.run(): #returns false, when target position is reached
        if self._stop == StopMode.HARDSTOP:
            break

    self._movement_phase = MovementPhase.STANDSTILL
    return self._stop



def run_to_position_revolutions(self, revolutions, movement_absolute_relative = None):
    """
    runs the motor to the given position.
    with acceleration and deceleration
    blocks the code until finished!

        Parameters:
            revolutions (int): amount of revs; can be negative
            movement_abs_rel (enum): whether the movement should be absolut or relative

        Returns:
            stop (enum): how the movement was finished
    """
    return self.run_to_position_steps(round(revolutions * self._steps_per_rev),
                                        movement_absolute_relative)



def run_to_position_steps_threaded(self, steps, movement_abs_rel = None):
    """
    runs the motor to the given position.
    with acceleration and deceleration
    does not block the code
    returns true when the movement if finshed normally and false,
    when the movement was stopped

        Parameters:
            steps (int): amount of steps; can be negative
            movement_abs_rel (enum): whether the movement should be absolut or relative

        Returns:
            stop (enum): how the movement was finished
    """
    self._movement_thread = threading.Thread(target=self.run_to_position_steps,
                                                args=(steps, movement_abs_rel))
    self._movement_thread.start()



def run_to_position_revolutions_threaded(self, revolutions, movement_absolute_relative = None):
    """
    runs the motor to the given position.
    with acceleration and deceleration
    does not block the code

        Parameters:
            revolutions (int): amount of revs; can be negative
            movement_abs_rel (enum): whether the movement should be absolut or relative

        Returns:
            stop (enum): how the movement was finished
    """
    return self.run_to_position_steps_threaded(round(revolutions * self._steps_per_rev),
                                                movement_absolute_relative)



def wait_for_movement_finished_threaded(self):
    """
    wait for the motor to finish the movement,
    if startet threaded
    returns true when the movement if finshed normally and false,
    when the movement was stopped

        Returns:
            stop (enum): how the movement was finished
    """
    self._movement_thread.join()
    return self._stop



def run(self):
    """
    calculates a new speed if a speed was made
    returns true if the target position is reached
    should not be called from outside!
    """
    if self.run_speed(): #returns true, when a step is made
        self.compute_new_speed()
    return self._speed != 0.0 and self.distance_to_go() != 0



def distance_to_go(self):
    """
    returns the remaining distance the motor should run
    """
    return self._target_pos - self._current_pos



def compute_new_speed(self):
    """
    returns the calculated current speed depending on the acceleration
    this code is based on:
    "Generate stepper-motor speed profiles in real time" by David Austin

    https://www.embedded.com/generate-stepper-motor-speed-profiles-in-real-time/
    https://web.archive.org/web/20140705143928/http://fab.cba.mit.edu/classes/MIT/961.09/projects/i0/Stepper_Motor_Speed_Profile.pdf
    """
    distance_to = self.distance_to_go() # +ve is clockwise from curent location
    steps_to_stop = (self._speed * self._speed) / (2.0 * self._acceleration) # Equation 16
    if ((distance_to == 0 and steps_to_stop <= 2) or
    (self._stop == StopMode.SOFTSTOP and steps_to_stop <= 1)):
        # We are at the target and its time to stop
        self._step_interval = 0
        self._speed = 0.0
        self._n = 0
        self._movement_phase = MovementPhase.STANDSTILL
        self.tmc_logger.log("time to stop", Loglevel.MOVEMENT)
        return

    if distance_to > 0:
        # We are anticlockwise from the target
        # Need to go clockwise from here, maybe decelerate now
        if self._n > 0:
            # Currently accelerating, need to decel now? Or maybe going the wrong way?
            if ((steps_to_stop >= distance_to) or self._direction == Direction.CCW or
                self._stop == StopMode.SOFTSTOP):
                self._n = -steps_to_stop # Start deceleration
                self._movement_phase = MovementPhase.DECELERATING
        elif self._n < 0:
            # Currently decelerating, need to accel again?
            if (steps_to_stop < distance_to) and self._direction == Direction.CW:
                self._n = -self._n # Start accceleration
                self._movement_phase = MovementPhase.ACCELERATING
    elif distance_to < 0:
        # We are clockwise from the target
        # Need to go anticlockwise from here, maybe decelerate
        if self._n > 0:
            # Currently accelerating, need to decel now? Or maybe going the wrong way?
            if (((steps_to_stop >= -distance_to) or self._direction == Direction.CW or
                self._stop == StopMode.SOFTSTOP)):
                self._n = -steps_to_stop # Start deceleration
                self._movement_phase = MovementPhase.DECELERATING
        elif self._n < 0:
            # Currently decelerating, need to accel again?
            if (steps_to_stop < -distance_to) and self._direction == Direction.CCW:
                self._n = -self._n # Start accceleration
                self._movement_phase = MovementPhase.ACCELERATING
    # Need to accelerate or decelerate
    if self._n == 0:
        # First step from stopped
        self._cn = self._c0
        GPIO.output(self._pin_step, GPIO.LOW)
        if distance_to > 0:
            self.set_direction_pin(1)
            self.tmc_logger.log("going CW", Loglevel.MOVEMENT)
        else:
            self.set_direction_pin(0)
            self.tmc_logger.log("going CCW", Loglevel.MOVEMENT)
        self._movement_phase = MovementPhase.ACCELERATING
    else:
        # Subsequent step. Works for accel (n is +_ve) and decel (n is -ve).
        self._cn = self._cn - ((2.0 * self._cn) / ((4.0 * self._n) + 1)) # Equation 13
        self._cn = max(self._cn, self._cmin)
        if self._cn == self._cmin:
            self._movement_phase = MovementPhase.MAXSPEED
    self._n += 1
    self._step_interval = self._cn
    self._speed = 1000000.0 / self._cn
    if self._direction == 0:
        self._speed = -self._speed



def run_speed(self):
    """
    this methods does the actual steps with the current speed
    """
    # Dont do anything unless we actually have a step interval
    if not self._step_interval:
        return False

    curtime = time.time_ns()/1000

    if curtime - self._last_step_time >= self._step_interval:

        if self._direction == 1: # Clockwise
            self._current_pos += 1
        else: # Anticlockwise
            self._current_pos -= 1
        self.make_a_step()

        self._last_step_time = curtime # Caution: does not account for costs in step()
        return True
    return False



def make_a_step(self):
    """
    method that makes on step
    for the TMC2209 there needs to be a signal duration of minimum 100 ns
    """
    GPIO.output(self._pin_step, GPIO.HIGH)
    time.sleep(1/1000/1000)
    GPIO.output(self._pin_step, GPIO.LOW)
    time.sleep(1/1000/1000)

    self.tmc_logger.log("one step", Loglevel.MOVEMENT)



#pylint: disable=invalid-name
#pylint: disable=import-error
#pylint: disable=broad-exception-caught
"""
TMC_UART stepper driver uart module
"""

import time
import struct
from bitstring import BitArray
import serial

from . import _TMC_2209_reg as reg


class TMC_UART:
    """
    TMC_UART

    this class is used to communicate with the TMC via UART
    it can be used to change the settings of the TMC.
    like the current or the microsteppingmode
    """
    tmc_logger = None

    mtr_id = 0
    ser = None
    r_frame  = [0x55, 0, 0, 0  ]
    w_frame  = [0x55, 0, 0, 0 , 0, 0, 0, 0 ]
    communication_pause = 0
    error_handler_running = False



    def __init__(self, tmc_logger, serialport, baudrate, mtr_id = 0):
        """
        constructor
        """
        self.tmc_logger = tmc_logger
        if serialport is None:
            return
        try:
            self.ser = serial.Serial (serialport, baudrate)
        except Exception as e:
            errnum = e.args[0]
            self.tmc_logger.log(f"SERIAL ERROR: {e}")
            if errnum == 2:
                self.tmc_logger.log(f""""{serialport} does not exist.
                      You need to activate the serial port with \"sudo raspi-config\"""")
            if errnum == 13:
                self.tmc_logger.log("""you have no permission to use the serial port.
                                    You may need to add your user to the dialout group
                                    with \"sudo usermod -a -G dialout pi\"""")

        self.mtr_id = mtr_id
        self.ser.BYTESIZES = 1
        self.ser.PARITIES = serial.PARITY_NONE
        self.ser.STOPBITS = 1

        # adjust per baud and hardware. Sequential reads without some delay fail.
        self.ser.timeout = 20000/baudrate
        # adjust per baud and hardware. Sequential reads without some delay fail.
        self.communication_pause = 500/baudrate


        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()



    def __del__(self):
        """
        destructor
        """
        if self.ser is not None:
            self.ser.close()



    def compute_crc8_atm(self, datagram, initial_value=0):
        """
        this function calculates the crc8 parity bit
        """
        crc = initial_value
        # Iterate bytes in data
        for byte in datagram:
            # Iterate bits in byte
            for _ in range(0, 8):
                if (crc >> 7) ^ (byte & 0x01):
                    crc = ((crc << 1) ^ 0x07) & 0xFF
                else:
                    crc = (crc << 1) & 0xFF
                # Shift to next bit
                byte = byte >> 1
        return crc



    def read_reg(self, register):
        """
        reads the registry on the TMC with a given address.
        returns the binary value of that register
        """

        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()

        self.r_frame[1] = self.mtr_id
        self.r_frame[2] = register
        self.r_frame[3] = self.compute_crc8_atm(self.r_frame[:-1])

        rtn = self.ser.write(self.r_frame)
        if rtn != len(self.r_frame):
            self.tmc_logger.log("Err in write")
            return False

        # adjust per baud and hardware. Sequential reads without some delay fail.
        time.sleep(self.communication_pause)

        rtn = self.ser.read(12)
        #self.tmc_logger.log(f"received {len(rtn)} bytes; {len(rtn*8)} bits")
        #self.tmc_logger.log(rtn.hex())

        time.sleep(self.communication_pause)

        return rtn



    def read_int(self, register, tries=10):
        """
        this function tries to read the registry of the TMC 10 times
        if a valid answer is returned, this function returns it as an integer
        """
        while True:
            tries -= 1
            rtn = self.read_reg(register)
            rtn_data = rtn[7:11]
            not_zero_count = len([elem for elem in rtn if elem != 0])

            if(len(rtn)<12 or not_zero_count == 0):
                self.tmc_logger.log(f"""UART Communication Error:
                                    {len(rtn_data)} data bytes | {len(rtn)} total bytes""")
            elif rtn[11] != self.compute_crc8_atm(rtn[4:11]):
                self.tmc_logger.log("UART Communication Error: CRC MISMATCH")
            else:
                break

            if tries<=0:
                self.tmc_logger.log("after 10 tries not valid answer")
                self.tmc_logger.log(f"snd:\t{bytes(self.r_frame)}")
                self.tmc_logger.log(f"rtn:\t{rtn}")
                self.handle_error()
                return -1

        val = struct.unpack(">i",rtn_data)[0]
        return val



    def write_reg(self, register, val):
        """
        this function can write a value to the register of the tmc
        1. use read_int to get the current setting of the TMC
        2. then modify the settings as wished
        3. write them back to the driver with this function
        """

        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()

        self.w_frame[1] = self.mtr_id
        self.w_frame[2] =  register | 0x80  # set write bit

        self.w_frame[3] = 0xFF & (val>>24)
        self.w_frame[4] = 0xFF & (val>>16)
        self.w_frame[5] = 0xFF & (val>>8)
        self.w_frame[6] = 0xFF & val

        self.w_frame[7] = self.compute_crc8_atm(self.w_frame[:-1])


        rtn = self.ser.write(self.w_frame)
        if rtn != len(self.w_frame):
            self.tmc_logger.log("Err in write")
            return False

        time.sleep(self.communication_pause)

        return True



    def write_reg_check(self, register, val, tries=10):
        """
        this function als writes a value to the register of the TMC
        but it also checks if the writing process was successfully by checking
        the InterfaceTransmissionCounter before and after writing
        """
        ifcnt1 = self.read_int(reg.IFCNT)

        if ifcnt1 == 255:
            ifcnt1 = -1

        while True:
            self.write_reg(register, val)
            tries -= 1
            ifcnt2 = self.read_int(reg.IFCNT)
            if ifcnt1 >= ifcnt2:
                self.tmc_logger.log("writing not successful!")
                self.tmc_logger.log("ifcnt:",ifcnt1,ifcnt2)
            else:
                return True
            if tries<=0:
                self.tmc_logger.log("after 10 tries no valid write access")
                self.handle_error()
                return -1



    def flush_serial_buffer(self):
        """
        this function clear the communication buffers of the Raspberry Pi
        """
        if self.ser is None:
            return
        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()



    def set_bit(self, value, bit):
        """
        this sets a specific bit to 1
        """
        return value | (bit)



    def clear_bit(self, value, bit):
        """
        this sets a specific bit to 0
        """
        return value & ~(bit)



    def handle_error(self):
        """
        error handling
        """
        if self.error_handler_running:
            return
        self.error_handler_running = True
        gstat = self.read_int(reg.GSTAT)
        self.tmc_logger.log("GSTAT Error check:")
        if gstat == -1:
            self.tmc_logger.log("No answer from Driver")
        elif gstat == 0:
            self.tmc_logger.log("Everything looks fine in GSTAT")
        else:
            if gstat & reg.reset:
                self.tmc_logger.log("The Driver has been reset since the last read access to GSTAT")
            if gstat & reg.drv_err:
                self.tmc_logger.log("""The driver has been shut down due to overtemperature or short
                      circuit detection since the last read access""")
            if gstat & reg.uv_cp:
                self.tmc_logger.log("""Undervoltage on the charge pump.
                      The driver is disabled in this case""")
        self.tmc_logger.log("EXITING!")
        raise SystemExit



    def test_uart(self, register):
        """
        test UART connection
        """

        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()

        self.r_frame[1] = self.mtr_id
        self.r_frame[2] = register
        self.r_frame[3] = self.compute_crc8_atm(self.r_frame[:-1])

        rtn = self.ser.write(self.r_frame)
        if rtn != len(self.r_frame):
            self.tmc_logger.log("Err in write")
            return False

        # adjust per baud and hardware. Sequential reads without some delay fail.
        time.sleep(self.communication_pause)

        rtn = self.ser.read(12)
        self.tmc_logger.log(f"received {len(rtn)} bytes; {len(rtn)*8} bits")
        self.tmc_logger.log(f"hex: {rtn.hex()}")
        c = BitArray(hex=rtn.hex())
        self.tmc_logger.log(f"bin: {c.bin}")

        time.sleep(self.communication_pause)

        return bytes(self.r_frame), rtn



#pylint: disable=invalid-name
"""
TMC_2209 stepper driver logger module
"""

from enum import Enum



class Loglevel(Enum):
    """loglevel"""
    NONE = 0
    ERROR = 10
    INFO = 20
    DEBUG = 30
    MOVEMENT = 40
    ALL = 100



class TMC_logger:
    """
    TMC_2209_comm

    this class has the function:
    move the motor via STEP/DIR pins
    """
    _loglevel = Loglevel.INFO
    _logprefix = "TMC2209"



    def __init__(self, loglevel = Loglevel.INFO, logprefix = "TMC2209"):
        """
        constructor

            Parameters:
                logprefix (string): new logprefix
                loglevel (enum): level for which to log
        """
        if loglevel is not None:
            self._loglevel = loglevel
        if logprefix is not None:
            self._logprefix = logprefix
        else:
            self._logprefix = "TMC2209"



    def set_logprefix(self, logprefix):
        """
        set the logprefix.

            Parameters:
                logprefix (string): new logprefix
        """
        self._logprefix = logprefix



    def set_loglevel(self, loglevel):
        """
        set the loglevel. See the Enum Loglevel

            Parameters:
                loglevel (enum): level for which to log
        """
        self._loglevel = loglevel



    def log(self, message, loglevel=Loglevel.NONE):
        """
        logs a message

            Parameters:
                message (string): message to log
                loglevel (enum): level for which to log
        """
        if self._loglevel.value >= loglevel.value:
            print(self._logprefix+": " +message)



#pylint: disable=invalid-name
#pylint: disable=missing-function-docstring
#pylint: disable=missing-class-docstring
"""
test for TMC_2209_math.py
"""

import unittest
import src.TMC_2209._TMC_2209_math as tmc_math

class TestTMCMath(unittest.TestCase):

    def test_rps_to_vactual(self):
        self.assertEqual(tmc_math.rps_to_vactual(1,400), 559, "rps_to_vactual is wrong")

    def test_vactual_to_rps(self):
        self.assertEqual(round(tmc_math.vactual_to_rps(559,400)), 1, "vactual_to_rps is wrong")

    def test_rps_to_steps(self):
        self.assertEqual(round(tmc_math.rps_to_steps(1,400)), 400, "rps_to_steps is wrong")

    def test_steps_to_rps(self):
        self.assertEqual(round(tmc_math.steps_to_rps(400,400)), 1, "steps_to_rps is wrong")

    def test_rps_to_tstep(self):
        self.assertEqual(round(tmc_math.rps_to_tstep(1,400, 2)), 234, "rps_to_tstep is wrong")

    def test_steps_to_tstep(self):
        self.assertEqual(round(tmc_math.steps_to_tstep(400,2)), 234, "steps_to_tstep is wrong")

if __name__ == '__main__':
    unittest.main()



#pylint: disable=invalid-name
#pylint: disable=too-many-public-methods
#pylint: disable=too-many-branches
#pylint: disable=protected-access
#pylint: disable=no-member
#pylint: disable=bare-except
"""
TMC_2209 stepper driver communication module
"""

import time
try:
    from RPi import GPIO
except:
    from Mock import GPIO
from ._TMC_2209_logger import Loglevel
from ._TMC_2209_move import MovementAbsRel, MovementPhase
from . import _TMC_2209_reg as tmc_reg



def test_dir_step_en(self):
    """
    tests the EN, DIR and STEP pin
    this sets the EN, DIR and STEP pin to HIGH, LOW and HIGH
    and checks the IOIN Register of the TMC meanwhile
    """
    pin_dir_ok = pin_step_ok = pin_en_ok = True

    GPIO.output(self._pin_step, GPIO.HIGH)
    GPIO.output(self._pin_dir, GPIO.HIGH)
    GPIO.output(self._pin_en, GPIO.HIGH)
    time.sleep(0.1)
    ioin = self.read_ioin()
    if not ioin & tmc_reg.io_dir:
        pin_dir_ok = False
    if not ioin & tmc_reg.io_step:
        pin_step_ok = False
    if not ioin & tmc_reg.io_enn:
        pin_en_ok = False

    GPIO.output(self._pin_step, GPIO.LOW)
    GPIO.output(self._pin_dir, GPIO.LOW)
    GPIO.output(self._pin_en, GPIO.LOW)
    time.sleep(0.1)
    ioin = self.read_ioin()
    if ioin & tmc_reg.io_dir:
        pin_dir_ok = False
    if ioin & tmc_reg.io_step:
        pin_step_ok = False
    if ioin & tmc_reg.io_enn:
        pin_en_ok = False

    GPIO.output(self._pin_step, GPIO.HIGH)
    GPIO.output(self._pin_dir, GPIO.HIGH)
    GPIO.output(self._pin_en, GPIO.HIGH)
    time.sleep(0.1)
    ioin = self.read_ioin()
    if not ioin & tmc_reg.io_dir:
        pin_dir_ok = False
    if not ioin & tmc_reg.io_step:
        pin_step_ok = False
    if not ioin & tmc_reg.io_enn:
        pin_en_ok = False

    self.set_motor_enabled(False)

    self.tmc_logger.log("---")
    if pin_dir_ok:
        self.tmc_logger.log("Pin DIR: \tOK")
    else:
        self.tmc_logger.log("Pin DIR: \tnot OK")
    if pin_step_ok:
        self.tmc_logger.log("Pin STEP: \tOK")
    else:
        self.tmc_logger.log("Pin STEP: \tnot OK")
    if pin_en_ok:
        self.tmc_logger.log("Pin EN: \tOK")
    else:
        self.tmc_logger.log("Pin EN: \tnot OK")
    self.tmc_logger.log("---")



def test_step(self):
    """
    test method
    """
    self.set_direction_pin(1)

    for _ in range(100):
        self._current_pos += 1
        GPIO.output(self._pin_step, GPIO.HIGH)
        time.sleep(0.001)
        GPIO.output(self._pin_step, GPIO.LOW)
        time.sleep(0.01)



def test_uart(self):
    """
    test method
    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("TEST UART")
    result = self.tmc_uart.test_uart(tmc_reg.IOIN)

    snd = result[0]
    rtn = result[1]

    self.tmc_logger.log(f"length snd: {len(snd)}", Loglevel.DEBUG)
    self.tmc_logger.log(f"length rtn: {len(rtn)}", Loglevel.DEBUG)

    if len(rtn)==12:
        self.tmc_logger.log("""the Raspberry Pi received the sended
                            bits and the answer from the TMC""",Loglevel.INFO)
    elif len(rtn)==4:
        self.tmc_logger.log("the Raspberry Pi received only the sended bits",
                            Loglevel.INFO)
    elif len(rtn)==0:
        self.tmc_logger.log("the Raspberry Pi did not receive anything",
                            Loglevel.INFO)
    else:
        self.tmc_logger.log(f"the Raspberry Pi received an unexpected amount of bits: {len(rtn)}",
                            Loglevel.INFO)

    if snd[0:4] == rtn[0:4]:
        self.tmc_logger.log("""the Raspberry Pi received exactly the bits it has send.
                    the first 4 bits are the same""", Loglevel.INFO)
    else:
        self.tmc_logger.log("""the Raspberry Pi did not received the bits it has send.
                    the first 4 bits are different""", Loglevel.INFO)


    self.tmc_logger.log("complete", Loglevel.DEBUG)
    self.tmc_logger.log(str(snd.hex()), Loglevel.DEBUG)
    self.tmc_logger.log(str(rtn.hex()), Loglevel.DEBUG)

    self.tmc_logger.log("just the first 4 bits", Loglevel.DEBUG)
    self.tmc_logger.log(str(snd[0:4].hex()), Loglevel.DEBUG)
    self.tmc_logger.log(str(rtn[0:4].hex()), Loglevel.DEBUG)


    self.tmc_logger.log("---")
    return True



def test_stallguard_threshold(self, steps):
    """
    test method for tuning stallguard threshold
    run this function with your motor settings and your motor load
    the function will determine the minimum stallguard results for each movement phase

        Parameters:
            steps (int): amount of steps; can be negative
    """

    self.tmc_logger.log("---", Loglevel.INFO)
    self.tmc_logger.log("test_stallguard_threshold", Loglevel.INFO)

    self.set_spreadcycle(0)

    min_stallguard_result_accel = 511
    min_stallguard_result_maxspeed = 511
    min_stallguard_result_decel = 511

    self.run_to_position_steps_threaded(steps, MovementAbsRel.RELATIVE)


    while self._movement_phase != MovementPhase.STANDSTILL:
        stallguard_result = self.get_stallguard_result()

        self.tmc_logger.log(f"{self._movement_phase} | {stallguard_result}",
                    Loglevel.INFO)

        if (self._movement_phase == MovementPhase.ACCELERATING and
            stallguard_result < min_stallguard_result_accel):
            min_stallguard_result_accel = stallguard_result
        if (self._movement_phase == MovementPhase.MAXSPEED and
            stallguard_result < min_stallguard_result_maxspeed):
            min_stallguard_result_maxspeed = stallguard_result
        if (self._movement_phase == MovementPhase.DECELERATING and
            stallguard_result < min_stallguard_result_decel):
            min_stallguard_result_decel = stallguard_result

    self.wait_for_movement_finished_threaded()

    self.tmc_logger.log("---", Loglevel.INFO)
    self.tmc_logger.log(f"min StallGuard result during accel: {min_stallguard_result_accel}",
                        Loglevel.INFO)
    self.tmc_logger.log(f"min StallGuard result during maxspeed: {min_stallguard_result_maxspeed}",
    Loglevel.INFO)
    self.tmc_logger.log(f"min StallGuard result during decel: {min_stallguard_result_decel}",
                        Loglevel.INFO)
    self.tmc_logger.log("---", Loglevel.INFO)



#pylint: disable=invalid-name
#pylint: disable=too-many-public-methods
#pylint: disable=too-many-branches
#pylint: disable=protected-access
"""
TMC_2209 stepper driver communication module
"""

import math
from ._TMC_2209_logger import Loglevel
from . import _TMC_2209_reg as tmc_reg



def read_drv_status(self):
    """
    read the register Adress "DRV_STATUS" and prints all current setting

        Returns:
            drvstatus (int): 32bit DRV_STATUS Register

    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("DRIVER STATUS:")
    drvstatus =self.tmc_uart.read_int(tmc_reg.DRVSTATUS)
    self.tmc_logger.log(bin(drvstatus), Loglevel.INFO)
    if drvstatus & tmc_reg.stst:
        self.tmc_logger.log("Info: motor is standing still")
    else:
        self.tmc_logger.log("Info: motor is running")

    if drvstatus & tmc_reg.stealth:
        self.tmc_logger.log("Info: motor is running on StealthChop")
    else:
        self.tmc_logger.log("Info: motor is running on SpreadCycle")

    cs_actual = drvstatus & tmc_reg.cs_actual
    cs_actual = cs_actual >> 16
    self.tmc_logger.log(f"CS actual: {cs_actual}")

    if drvstatus & tmc_reg.olb:
        self.tmc_logger.log("Warning: Open load detected on phase B")

    if drvstatus & tmc_reg.ola:
        self.tmc_logger.log("Warning: Open load detected on phase A")

    if drvstatus & tmc_reg.s2vsb:
        self.tmc_logger.log("""Error: Short on low-side MOSFET detected on phase B.
                    The driver becomes disabled""")

    if drvstatus & tmc_reg.s2vsa:
        self.tmc_logger.log("""Error: Short on low-side MOSFET detected on phase A.
                    The driver becomes disabled""")

    if drvstatus & tmc_reg.s2gb:
        self.tmc_logger.log("""Error: Short to GND detected on phase B.
                            The driver becomes disabled.""")

    if drvstatus & tmc_reg.s2ga:
        self.tmc_logger.log("""Error: Short to GND detected on phase A.
                            The driver becomes disabled.""")

    if drvstatus & tmc_reg.ot:
        self.tmc_logger.log("Error: Driver Overheating!")

    if drvstatus & tmc_reg.otpw:
        self.tmc_logger.log("Warning: Driver Overheating Prewarning!")

    self.tmc_logger.log("---")
    return drvstatus



def read_gconf(self):
    """
    read the register Adress "GCONF" and prints all current setting

        Returns:
            gconf (int): 10bit GCONF Register
    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("GENERAL CONFIG")
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    self.tmc_logger.log(bin(gconf), Loglevel.INFO)

    if gconf & tmc_reg.i_scale_analog:
        self.tmc_logger.log("Driver is using voltage supplied to VREF as current reference")
    else:
        self.tmc_logger.log("Driver is using internal reference derived from 5VOUT")
    if gconf & tmc_reg.internal_rsense:
        self.tmc_logger.log("""Internal sense resistors.
                            Use current supplied into VREF as reference.""")
        self.tmc_logger.log("VREF pin internally is driven to GND in this mode.")
        self.tmc_logger.log("This will most likely destroy your driver!!!")
        raise SystemExit
    self.tmc_logger.log("Operation with external sense resistors")
    if gconf & tmc_reg.en_spreadcycle:
        self.tmc_logger.log("SpreadCycle mode enabled")
    else:
        self.tmc_logger.log("StealthChop PWM mode enabled")
    if gconf & tmc_reg.shaft:
        self.tmc_logger.log("Inverse motor direction")
    else:
        self.tmc_logger.log("normal motor direction")
    if gconf & tmc_reg.index_otpw:
        self.tmc_logger.log("INDEX pin outputs overtemperature prewarning flag")
    else:
        self.tmc_logger.log("INDEX shows the first microstep position of sequencer")
    if gconf & tmc_reg.index_step:
        self.tmc_logger.log("INDEX output shows step pulses from internal pulse generator")
    else:
        self.tmc_logger.log("INDEX output as selected by index_otpw")
    if gconf & tmc_reg.mstep_reg_select:
        self.tmc_logger.log("Microstep resolution selected by MSTEP register")
    else:
        self.tmc_logger.log("Microstep resolution selected by pins MS1, MS2")

    self.tmc_logger.log("---")
    return gconf



def read_gstat(self):
    """
    read the register Adress "GSTAT" and prints all current setting

        Returns:
            gstat (int): 3bit GSTAT Register
    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("GSTAT")
    gstat = self.tmc_uart.read_int(tmc_reg.GSTAT)
    self.tmc_logger.log(bin(gstat), Loglevel.INFO)
    if gstat & tmc_reg.reset:
        self.tmc_logger.log("The Driver has been reset since the last read access to GSTAT")
    if gstat & tmc_reg.drv_err:
        self.tmc_logger.log("""The driver has been shut down due to overtemperature or
                    short circuit detection since the last read access""")
    if gstat & tmc_reg.uv_cp:
        self.tmc_logger.log("""Undervoltage on the charge pump.
                            The driver is disabled in this case""")
    self.tmc_logger.log("---")
    return gstat



def clear_gstat(self):
    """
    read the register Adress "GSTAT" and prints all current setting
    """
    self.tmc_logger.log("clearing GSTAT", Loglevel.INFO)
    gstat = self.tmc_uart.read_int(tmc_reg.GSTAT)

    gstat = self.tmc_uart.set_bit(gstat, tmc_reg.reset)
    gstat = self.tmc_uart.set_bit(gstat, tmc_reg.drv_err)

    self.tmc_uart.write_reg_check(tmc_reg.GSTAT, gstat)



def read_ioin(self):
    """
    read the register Adress "IOIN" and prints all current setting

        Returns:
            ioin (int): 10+8bit IOIN Register
    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("INPUTS")
    ioin = self.tmc_uart.read_int(tmc_reg.IOIN)
    self.tmc_logger.log(bin(ioin), Loglevel.INFO)
    if ioin & tmc_reg.io_spread:
        self.tmc_logger.log("spread is high")
    else:
        self.tmc_logger.log("spread is low")

    if ioin & tmc_reg.io_dir:
        self.tmc_logger.log("dir is high")
    else:
        self.tmc_logger.log("dir is low")

    if ioin & tmc_reg.io_step:
        self.tmc_logger.log("step is high")
    else:
        self.tmc_logger.log("step is low")

    if ioin & tmc_reg.io_enn:
        self.tmc_logger.log("en is high")
    else:
        self.tmc_logger.log("en is low")

    self.tmc_logger.log("---")
    return ioin



def read_chopconf(self):
    """
    read the register Adress "CHOPCONF" and prints all current setting

        Returns:
            chopconf (int): 3bit CHOPCONF Register
    """
    self.tmc_logger.log("---")
    self.tmc_logger.log("CHOPPER CONTROL")
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)
    self.tmc_logger.log(bin(chopconf), Loglevel.INFO)

    self.tmc_logger.log(f"native {self.get_microstepping_resolution()} microstep setting")

    if chopconf & tmc_reg.intpol:
        self.tmc_logger.log("interpolation to 256 microsteps")

    if chopconf & tmc_reg.vsense:
        self.tmc_logger.log("1: High sensitivity, low sense resistor voltage")
    else:
        self.tmc_logger.log("0: Low sensitivity, high sense resistor voltage")

    self.tmc_logger.log("---")
    return chopconf



def get_direction_reg(self):
    """
    returns the motor shaft direction: 0 = CCW; 1 = CW

        Returns:
            direction (bool): motor shaft direction: 0 = CCW; 1 = CW
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    return gconf & tmc_reg.shaft



def set_direction_reg(self, direction):
    """
    sets the motor shaft direction to the given value: 0 = CCW; 1 = CW

        Parameters:
            direction (bool): motor shaft direction: 0 = CCW; 1 = CW
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    if direction:
        self.tmc_logger.log("write inverse motor direction", Loglevel.INFO)
        gconf = self.tmc_uart.set_bit(gconf, tmc_reg.shaft)
    else:
        self.tmc_logger.log("write normal motor direction", Loglevel.INFO)
        gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.shaft)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def get_iscale_analog(self):
    """
    return whether Vref (1) or 5V (0) is used for current scale

        Returns:
            en (bool): whether Vref (1) or 5V (0) is used for current scale
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    return gconf & tmc_reg.i_scale_analog



def set_iscale_analog(self,en):
    """
    sets Vref (1) or 5V (0) for current scale

        Parameters:
            en (bool): whether Vref (1) or 5V (0) is used for current scale
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    if en:
        self.tmc_logger.log("activated Vref for current scale", Loglevel.INFO)
        gconf = self.tmc_uart.set_bit(gconf, tmc_reg.i_scale_analog)
    else:
        self.tmc_logger.log("activated 5V-out for current scale", Loglevel.INFO)
        gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.i_scale_analog)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def get_vsense(self):
    """
    returns which sense resistor voltage is used for current scaling
    0: Low sensitivity, high sense resistor voltage
    1: High sensitivity, low sense resistor voltage

        Returns:
            en (bool): whether high sensitivity should is used
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)
    return chopconf & tmc_reg.vsense



def set_vsense(self,en):
    """
    sets which sense resistor voltage is used for current scaling
    0: Low sensitivity, high sense resistor voltage
    1: High sensitivity, low sense resistor voltage

        Parameters:
            en (bool): whether high sensitivity should be used
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)
    if en:
        self.tmc_logger.log("activated High sensitivity, low sense resistor voltage",
                            Loglevel.INFO)
        chopconf = self.tmc_uart.set_bit(chopconf, tmc_reg.vsense)
    else:
        self.tmc_logger.log("activated Low sensitivity, high sense resistor voltage",
                            Loglevel.INFO)
        chopconf = self.tmc_uart.clear_bit(chopconf, tmc_reg.vsense)
    self.tmc_uart.write_reg_check(tmc_reg.CHOPCONF, chopconf)



def get_internal_rsense(self):
    """
    returns which sense resistor voltage is used for current scaling
    0: Operation with external sense resistors
    1: Internal sense resistors. Use current supplied into
    VREF as reference for internal sense resistor. VREF
    pin internally is driven to GND in this mode.

        Returns:
            en (bool): which sense resistor voltage is used
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    return gconf & tmc_reg.internal_rsense



def set_internal_rsense(self,en):
    """
    sets which sense resistor voltage is used for current scaling
    0: Operation with external sense resistors
    1: Internal sense resistors. Use current supplied into
    VREF as reference for internal sense resistor. VREF
    pin internally is driven to GND in this mode.

        Parameters:
            en (bool):which sense resistor voltage is used; true will propably destroy your tmc
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    if en:
        self.tmc_logger.log("activated internal sense resistors.",
                            Loglevel.INFO)
        self.tmc_logger.log("VREF pin internally is driven to GND in this mode.",
                            Loglevel.INFO)
        self.tmc_logger.log("This will most likely destroy your driver!!!",
                            Loglevel.INFO)
        raise SystemExit
        # gconf = self.tmc_uart.set_bit(gconf, tmc_reg.internal_rsense)
    self.tmc_logger.log("activated operation with external sense resistors",
                        Loglevel.INFO)
    gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.internal_rsense)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def set_irun_ihold(self, ihold, irun, ihold_delay):
    """
    sets the current scale (CS) for Running and Holding
    and the delay, when to be switched to Holding current

        Parameters:
            ihold (int): 0-31
            irun (int): 0-31
            ihold_delay (int): 0-15
    """
    ihold_irun = 0

    ihold_irun = ihold_irun | ihold << 0
    ihold_irun = ihold_irun | irun << 8
    ihold_irun = ihold_irun | ihold_delay << 16
    self.tmc_logger.log(f"ihold_irun: {bin(ihold_irun)}", Loglevel.INFO)

    self.tmc_logger.log("writing ihold_irun", Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.IHOLD_IRUN, ihold_irun)



def set_pdn_disable(self,pdn_disable):
    """
    disables PDN on the UART pin
    0: PDN_UART controls standstill current reduction
    1: PDN_UART input function disabled. Set this bit,
    when using the UART interface!

        Parameters:
            pdn_disable (bool): whether pdn should be disabled
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    if pdn_disable:
        self.tmc_logger.log("enabled PDN_UART", Loglevel.INFO)
        gconf = self.tmc_uart.set_bit(gconf, tmc_reg.pdn_disable)
    else:
        self.tmc_logger.log("disabled PDN_UART", Loglevel.INFO)
        gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.pdn_disable)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def set_current(self, run_current, hold_current_multiplier = 0.5,
                hold_current_delay = 10, pdn_disable = True):
    """
    sets the current flow for the motor.

        Parameters:
            run_current (int): current during movement in mA
            hold_current_multiplier (int): current multiplier during standstill
            hold_current_delay (int): delay after standstill after which the current drops
            use_vref (int): most driver have a trimpotentiometer to set the vref. optional
            Vref (int): if you want to use Vref, you need to set the Vref Voltage
            pdn_disable (bool): should be disabled if UART is used
    """
    cs_irun = 0
    rsense = 0.11
    vfs = 0

    self.set_iscale_analog(False)

    vfs = 0.325
    cs_irun = 32.0*1.41421*run_current/1000.0*(rsense+0.02)/vfs - 1

    # If Current Scale is too low, turn on high sensitivity VSsense and calculate again
    if cs_irun < 16:
        self.tmc_logger.log("CS too low; switching to VSense True", Loglevel.INFO)
        vfs = 0.180
        cs_irun = 32.0*1.41421*run_current/1000.0*(rsense+0.02)/vfs - 1
        self.set_vsense(True)
    else: # If CS >= 16, turn off high_senser
        self.tmc_logger.log("CS in range; using VSense False", Loglevel.INFO)
        self.set_vsense(False)

    cs_irun = min(cs_irun, 31)
    cs_irun = max(cs_irun, 0)

    CS_IHold = hold_current_multiplier * cs_irun

    cs_irun = round(cs_irun)
    CS_IHold = round(CS_IHold)
    hold_current_delay = round(hold_current_delay)

    self.tmc_logger.log(f"cs_irun: {cs_irun}", Loglevel.INFO)
    self.tmc_logger.log(f"CS_IHold: {CS_IHold}", Loglevel.INFO)
    self.tmc_logger.log(f"Delay: {hold_current_delay}", Loglevel.INFO)

    # return (float)(CS+1)/32.0 * (vsense() ? 0.180 : 0.325)/(rsense+0.02) / 1.41421 * 1000;
    run_current_actual = (cs_irun+1)/32.0 * (vfs)/(rsense+0.02) / 1.41421 * 1000
    self.tmc_logger.log(f"actual current: {round(run_current_actual)} mA",
                        Loglevel.INFO)

    self.set_irun_ihold(CS_IHold, cs_irun, hold_current_delay)

    self.set_pdn_disable(pdn_disable)



def get_spreadcycle(self):
    """
    return whether spreadcycle (1) is active or stealthchop (0)

        Parameters:
            en_spread (bool): true = spreadcycle; false = stealthchop
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    return gconf & tmc_reg.en_spreadcycle



def set_spreadcycle(self,en_spread):
    """
    enables spreadcycle (1) or stealthchop (0)

        Parameters:
            en_spread (bool): true to enable spreadcycle; false to enable stealthchop
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)
    if en_spread:
        self.tmc_logger.log("activated Spreadcycle", Loglevel.INFO)
        gconf = self.tmc_uart.set_bit(gconf, tmc_reg.en_spreadcycle)
    else:
        self.tmc_logger.log("activated Stealthchop", Loglevel.INFO)
        gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.en_spreadcycle)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def get_interpolation(self):
    """
    return whether the tmc inbuilt interpolation is active

        Returns:
            en (bool): true if internal µstep interpolation is enabled
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)
    return bool(chopconf & tmc_reg.intpol)



def set_interpolation(self, en):
    """
    enables the tmc inbuilt interpolation of the steps to 256 microsteps

        Parameters:
            en (bool): true to enable internal µstep interpolation
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)

    if en:
        chopconf = self.tmc_uart.set_bit(chopconf, tmc_reg.intpol)
    else:
        chopconf = self.tmc_uart.clear_bit(chopconf, tmc_reg.intpol)

    self.tmc_logger.log(f"writing microstep interpolation setting: {str(en)}",
                        Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.CHOPCONF, chopconf)



def read_microstepping_resolution(self):
    """
    returns the current native microstep resolution (1-256)
    this reads the value from the driver register

        Returns:
            msres (int): µstep resolution
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)

    msresdezimal = chopconf & (tmc_reg.msres0 | tmc_reg.msres1 |
                                tmc_reg.msres2 | tmc_reg.msres3)

    msresdezimal = msresdezimal >> 24
    msresdezimal = 8 - msresdezimal

    self._msres = int(math.pow(2, msresdezimal))

    return self._msres



def get_microstepping_resolution(self):
    """
    returns the current native microstep resolution (1-256)
    this returns the cached value from this module

        Returns:
            msres (int): µstep resolution
    """
    return self._msres



def set_microstepping_resolution(self, msres):
    """
    sets the current native microstep resolution (1,2,4,8,16,32,64,128,256)

        Parameters:
            msres (int): µstep resolution; has to be a power of 2 or 1 for fullstep
    """
    chopconf = self.tmc_uart.read_int(tmc_reg.CHOPCONF)
    #setting all bits to zero
    chopconf = chopconf & (~tmc_reg.msres0 | ~tmc_reg.msres1 |
                            ~tmc_reg.msres2 | ~tmc_reg.msres3)
    msresdezimal = int(math.log(msres, 2))
    msresdezimal = 8 - msresdezimal
    chopconf = int(chopconf) & int(4043309055)
    chopconf = chopconf | msresdezimal <<24

    self.tmc_logger.log(f"writing {msres} microstep setting", Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.CHOPCONF, chopconf)

    self._msres = msres

    self.set_mstep_resolution_reg_select(True)

    return True



def set_mstep_resolution_reg_select(self, en):
    """
    sets the register bit "mstep_reg_select" to 1 or 0 depending to the given value.
    this is needed to set the microstep resolution via UART
    this method is called by "set_microstepping_resolution"

        Parameters:
            en (bool): true to set µstep resolution via UART
    """
    gconf = self.tmc_uart.read_int(tmc_reg.GCONF)

    if en is True:
        gconf = self.tmc_uart.set_bit(gconf, tmc_reg.mstep_reg_select)
    else:
        gconf = self.tmc_uart.clear_bit(gconf, tmc_reg.mstep_reg_select)

    self.tmc_logger.log(f"writing MStep Reg Select: {en}", Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.GCONF, gconf)



def get_interface_transmission_counter(self):
    """
    reads the interface transmission counter from the tmc register
    this value is increased on every succesfull write access
    can be used to verify a write access

        Returns:
            ifcnt (int): 8bit IFCNT Register
    """
    ifcnt = self.tmc_uart.read_int(tmc_reg.IFCNT)
    self.tmc_logger.log(f"Interface Transmission Counter: {ifcnt}", Loglevel.INFO)
    return ifcnt



def get_tstep(self):
    """
    reads the current tstep from the driver register

        Returns:
            tstep (int): TStep time
    """
    tstep = self.tmc_uart.read_int(tmc_reg.TSTEP)
    return tstep



def set_vactual(self, vactual):
    """
    sets the register bit "VACTUAL" to to a given value
    VACTUAL allows moving the motor by UART control.
    It gives the motor velocity in +-(2^23)-1 [μsteps / t]
    0: Normal operation. Driver reacts to STEP input

        Parameters:
            vactual (int): value for vactual
    """
    self.tmc_uart.write_reg_check(tmc_reg.VACTUAL, vactual)



def get_stallguard_result(self):
    """
    return the current stallguard result
    its will be calculated with every fullstep
    higher values means a lower motor load

        Returns:
            sg_result (int): StallGuard Result
    """
    sg_result = self.tmc_uart.read_int(tmc_reg.SG_RESULT)
    return sg_result



def set_stallguard_threshold(self, threshold):
    """
    sets the register bit "SGTHRS" to to a given value
    this is needed for the stallguard interrupt callback
    SG_RESULT becomes compared to the double of this threshold.
    SG_RESULT ≤ SGTHRS*2

        Parameters:
            threshold (int): value for SGTHRS
    """

    self.tmc_logger.log(f"sgthrs {bin(threshold)}", Loglevel.INFO)

    self.tmc_logger.log("writing sgthrs", Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.SGTHRS, threshold)



def set_coolstep_threshold(self, threshold):
    """
    This  is  the  lower  threshold  velocity  for  switching
    on  smart energy CoolStep and StallGuard to DIAG output. (unsigned)

        Parameters:
            threshold (int): threshold  velocity
    """

    self.tmc_logger.log(f"tcoolthrs {bin(threshold)}", Loglevel.INFO)

    self.tmc_logger.log("writing tcoolthrs", Loglevel.INFO)
    self.tmc_uart.write_reg_check(tmc_reg.TCOOLTHRS, threshold)



def get_microstep_counter(self):
    """
    returns the current Microstep counter.
    Indicates actual position in the microstep table for CUR_A

        Returns:
            mscnt (int): current Microstep counter
    """
    mscnt = self.tmc_uart.read_int(tmc_reg.MSCNT)
    return mscnt



def get_microstep_counter_in_steps(self, offset=0):
    """
    returns the current Microstep counter.
    Indicates actual position in the microstep table for CUR_A

        Parameters:
            offset (int): offset added to the current step counter

        Returns:
            step (int): current Microstep counter convertet to steps
    """
    step = (self.get_microstep_counter()-64)*(self._msres*4)/1024
    step = (4*self._msres)-step-1
    step = round(step)
    return step+offset



#pylint: disable=invalid-name
#pylint: disable=missing-function-docstring
#pylint: disable=missing-class-docstring
#pylint: disable=wildcard-import
#pylint: disable=pointless-statement
#pylint: disable=unused-wildcard-import
"""
test for TMC_2209_uart.py
"""

import unittest
import mock
from src.TMC_2209._TMC_2209_logger import *
from src.TMC_2209._TMC_2209_uart import *


class TestTMCUart(unittest.TestCase):

    def setUp(self):
        self.tmc_logger = TMC_logger()
        self.tmc_uart = TMC_UART(self.tmc_logger, serialport=None, baudrate=115200)

    def test_read_int(self):
        with mock.patch.object(TMC_UART, 'read_reg', return_value=
                               b'U\x00o\x03\x05\xffo\xc0\x1e\x00\x00\xca'):
            reg_ans = self.tmc_uart.read_int(0x00)
            self.assertEqual(reg_ans, -1071775744, "read_int is wrong")


if __name__ == '__main__':
    unittest.main()



#pylint: disable=invalid-name
"""
this file contains:
1. hexadecimal address of the different registers
2. bitposition and bitmasks of the different values of each register

Example:
the register IOIN has the address 0x06 and the first bit shows
whether the ENABLE (EN/ENN) Pin is currently HIGH or LOW
"""

#addresses
GCONF           =   0x00
GSTAT           =   0x01
IFCNT           =   0x02
IOIN            =   0x06
IHOLD_IRUN      =   0x10
TSTEP           =   0x12
VACTUAL         =   0x22
TCOOLTHRS       =   0x14
SGTHRS          =   0x40
SG_RESULT       =   0x41
MSCNT           =   0x6A
CHOPCONF        =   0x6C
DRVSTATUS       =   0x6F

#GCONF
i_scale_analog      = 1<<0
internal_rsense     = 1<<1
en_spreadcycle      = 1<<2
shaft               = 1<<3
index_otpw          = 1<<4
index_step          = 1<<5
pdn_disable         = 1<<6
mstep_reg_select    = 1<<7

#GSTAT
reset               = 1<<0
drv_err             = 1<<1
uv_cp               = 1<<2

#CHOPCONF
vsense              = 1<<17
msres0              = 1<<24
msres1              = 1<<25
msres2              = 1<<26
msres3              = 1<<27
intpol              = 1<<28

#IOIN
io_enn              = 1<<0
io_step             = 1<<7
io_spread           = 1<<8
io_dir              = 1<<9

#DRVSTATUS
stst                = 1<<31
stealth             = 1<<30
cs_actual           = 31<<16
t157                = 1<<11
t150                = 1<<10
t143                = 1<<9
t120                = 1<<8
olb                 = 1<<7
ola                 = 1<<6
s2vsb               = 1<<5
s2vsa               = 1<<4
s2gb                = 1<<3
s2ga                = 1<<2
ot                  = 1<<1
otpw                = 1<<0

#IHOLD_IRUN
ihold               = 31<<0
irun                = 31<<8
iholddelay          = 15<<16

#SGTHRS
sgthrs              = 255<<0

#others
mres_256 = 0
mres_128 = 1
mres_64 = 2
mres_32 = 3
mres_16 = 4
mres_8 = 5
mres_4 = 6
mres_2 = 7
mres_1 = 8



#pylint: disable=invalid-name
#pylint: disable=no-member
#pylint: disable=too-many-arguments
#pylint: disable=too-many-public-methods
#pylint: disable=too-many-branches
#pylint: disable=too-many-instance-attributes
#pylint: disable=import-outside-toplevel
#pylint: disable=bare-except
"""
TMC_2209 stepper driver module
"""

import time
import statistics
try:
    from RPi import GPIO
except:
    from Mock import GPIO
from ._TMC_2209_uart import TMC_UART as tmc_uart
from ._TMC_2209_logger import TMC_logger, Loglevel
from ._TMC_2209_move import MovementAbsRel, MovementPhase, StopMode
from . import _TMC_2209_math as tmc_math




class TMC_2209:
    """
    TMC_2209

    this class has two different functions:
    1. change setting in the TMC-driver via UART
    2. move the motor via STEP/DIR pins
    """

    from ._TMC_2209_comm import (
        read_drv_status, read_gconf, read_gstat, clear_gstat, read_ioin, read_chopconf,
        get_direction_reg, set_direction_reg, get_iscale_analog, set_iscale_analog,get_vsense,
        set_vsense, get_internal_rsense, set_internal_rsense, set_irun_ihold, set_pdn_disable,
        set_current, get_spreadcycle, set_spreadcycle, get_interpolation, set_interpolation,
        read_microstepping_resolution, get_microstepping_resolution, set_microstepping_resolution,
        set_mstep_resolution_reg_select, get_interface_transmission_counter, get_tstep, set_vactual,
        get_stallguard_result, set_stallguard_threshold, set_coolstep_threshold,
        get_microstep_counter, get_microstep_counter_in_steps
    )

    from ._TMC_2209_move import (
        set_movement_abs_rel, get_current_position, set_current_position, set_max_speed,
        set_max_speed_fullstep, get_max_speed, set_acceleration, set_acceleration_fullstep,
        get_acceleration, stop, get_movement_phase, run_to_position_steps,
        run_to_position_revolutions, run_to_position_steps_threaded,
        run_to_position_revolutions_threaded, wait_for_movement_finished_threaded, run,
        distance_to_go, compute_new_speed, run_speed, make_a_step
    )

    from ._TMC_2209_test import (
        test_dir_step_en, test_step, test_uart, test_stallguard_threshold
    )

    tmc_uart = None
    tmc_logger = None
    _pin_step = -1
    _pin_dir = -1
    _pin_en = -1
    _pin_stallguard = -1

    _direction = True

    _stop = StopMode.NO
    _starttime = 0
    _sg_delay = 0
    _sg_callback = None

    _msres = -1
    _steps_per_rev = 0
    _fullsteps_per_rev = 200

    _current_pos = 0                 # current position of stepper in steps
    _target_pos = 0                  # the target position in steps
    _speed = 0.0                    # the current speed in steps per second
    _max_speed = 1.0                 # the maximum speed in steps per second
    _max_speed_homing = 200           # the maximum speed in steps per second for homing
    _acceleration = 1.0             # the acceleration in steps per second per second
    _acceleration_homing = 10000     # the acceleration in steps per second per second for homing
    _sqrt_twoa = 1.0                # Precomputed sqrt(2*_acceleration)
    _step_interval = 0               # the current interval between two steps
    _min_pulse_width = 1              # minimum allowed pulse with in microseconds
    _last_step_time = 0               # The last step time in microseconds
    _n = 0                          # step counter
    _c0 = 0                         # Initial step size in microseconds
    _cn = 0                         # Last step size in microseconds
    _cmin = 0                       # Min step size in microseconds based on maxSpeed
    _sg_threshold = 100             # threshold for stallguard
    _movement_abs_rel = MovementAbsRel.ABSOLUTE
    _movement_phase = MovementPhase.STANDSTILL

    _movement_thread = None

    _deinit_finished = False



    def __init__(self, pin_en, pin_step=-1, pin_dir=-1, baudrate=115200, serialport="/dev/serial0",
                 driver_address=0, gpio_mode=GPIO.BCM, loglevel=None, skip_uart_init=False):
        """
        constructor
        """
        self.init(pin_en, pin_step, pin_dir, baudrate, serialport, driver_address,
                  gpio_mode, loglevel, skip_uart_init)



    def __del__(self):
        """
        destructor
        """
        self.deinit()
        del self.tmc_uart
        del self.tmc_logger



    def init(self, pin_en, pin_step=-1, pin_dir=-1, baudrate=115200, serialport="/dev/serial0",
    driver_address=0, gpio_mode=GPIO.BCM, loglevel=None, skip_uart_init=False):
        """
        init function

            Parameters:
                pin_en (int): Pin number of EN pin
                pin_step (int): Pin number of STEP pin
                pin_dir (int): Pin number of DIR pin
                baudrate (int): baudrate exp: 9600, 115200
                serialport (string): win: 'COM1'; linux: 'dev/serial0'
                driver_address (int): 0 - 3
                no_uart=False (bool): skip UART init, if only STEP/DIR is used
                gpio_mode (enum): numbering system for the GPIO pins
                loglevel (enum): level for which to log
        """
        self.tmc_logger = TMC_logger(loglevel, f"TMC2209 {driver_address}")
        self.tmc_uart = tmc_uart(self.tmc_logger, serialport, baudrate, driver_address)


        self.tmc_logger.log("Init", Loglevel.INFO)
        GPIO.setwarnings(False)
        GPIO.setmode(gpio_mode)

        self.tmc_logger.log("EN Pin: {pin_en}", Loglevel.DEBUG)
        self._pin_en = pin_en
        GPIO.setup(self._pin_en, GPIO.OUT, initial=GPIO.HIGH)

        self.tmc_logger.log("STEP Pin: {pin_step}", Loglevel.DEBUG)
        if pin_step != -1:
            self._pin_step = pin_step
            GPIO.setup(self._pin_step, GPIO.OUT, initial=GPIO.LOW)

        self.tmc_logger.log("DIR Pin: {pin_dir}", Loglevel.DEBUG)
        if pin_dir != -1:
            self._pin_dir = pin_dir
            GPIO.setup(self._pin_dir, GPIO.OUT, initial=self._direction)

        self.tmc_logger.log("GPIO Init finished", Loglevel.INFO)

        if not skip_uart_init:
            self.read_steps_per_rev()
            self.clear_gstat()

        self.tmc_uart.flush_serial_buffer()
        self.tmc_logger.log("Init finished", Loglevel.INFO)

        self.set_max_speed_fullstep(100)
        self.set_acceleration_fullstep(100)



    def deinit(self):
        """
        deinit function
        """
        if self._deinit_finished is False:
            self.tmc_logger.log("Deinit", Loglevel.INFO)

            self.set_motor_enabled(False)

            self.tmc_logger.log("GPIO cleanup")
            if self._pin_step != -1:
                GPIO.cleanup(self._pin_step)
            if self._pin_dir != -1:
                GPIO.cleanup(self._pin_dir)
            if self._pin_en != -1:
                GPIO.cleanup(self._pin_en)
            if self._pin_stallguard != -1:
                GPIO.remove_event_detect(self._pin_stallguard)
                GPIO.cleanup(self._pin_stallguard)

            self.tmc_logger.log("Deinit finished", Loglevel.INFO)
            self._deinit_finished= True
        else:
            self.tmc_logger.log("Deinit already finished", Loglevel.INFO)



    def set_deinitialize_true(self):
        """
        set deinitialize to true
        """
        self._deinit_finished = True



    def set_motor_enabled(self, en):
        """
        enables or disables the motor current output

            Parameters:
                en (bool): whether the motor current output should be enabled
        """
        GPIO.output(self._pin_en, not en)
        self.tmc_logger.log(f"Motor output active: {en}", Loglevel.INFO)



    def do_homing(self, diag_pin, revolutions = 10, threshold = None, speed_rpm = None):
        """
        homes the motor in the given direction using stallguard.
        this method is using vactual to move the motor and an interrupt on the DIAG pin

            Parameters
                diag_pin (int): DIAG pin number
                revolutions (int): max number of revolutions. Can be negative for inverse direction
                threshold (int): optional; StallGuard detection threshold
                speed_rpm (float): optional; speed in revolutions per minute

            Returns:
                not homing_failed (bool): true when homing was successful
        """
        if threshold is not None:
            self._sg_threshold = threshold
        if speed_rpm is None:
            speed_rpm = tmc_math.steps_to_rps(self._max_speed_homing, self._steps_per_rev)*60

        self.tmc_logger.log("---", Loglevel.INFO)
        self.tmc_logger.log("homing", Loglevel.INFO)

        # StallGuard only works with StealthChop
        self.set_spreadcycle(0)

        self.tmc_logger.log(f"Stallguard threshold: {self._sg_threshold}", Loglevel.DEBUG)

        self.set_stallguard_callback(diag_pin, self._sg_threshold, self.stop,
                                     0.5*tmc_math.rps_to_steps(speed_rpm/60, self._steps_per_rev))

        homing_failed = self.set_vactual_rpm(speed_rpm, revolutions=revolutions)

        if homing_failed:
            self.tmc_logger.log("homing failed", Loglevel.INFO)
        else:
            self.tmc_logger.log("homing successful",Loglevel.INFO)

        self._current_pos = 0

        self.tmc_logger.log("---", Loglevel.INFO)
        return not homing_failed



    def do_homing2(self, revolutions, threshold=None):
        """
        homes the motor in the given direction using stallguard
        old function, uses STEP/DIR to move the motor and pulls the StallGuard result
        from the interface

            Parameters
                revolutions (int): max number of revolutions. Can be negative for inverse direction
                threshold (int): optional; StallGuard detection threshold
        """
        sg_results = []

        if threshold is not None:
            self._sg_threshold = threshold

        self.tmc_logger.log("---", Loglevel.INFO)
        self.tmc_logger.log("homing", Loglevel.INFO)

        self.tmc_logger.log(f"Stallguard threshold: {self._sg_threshold}", Loglevel.DEBUG)

        self.set_direction_pin(revolutions > 0)

        # StallGuard only works with StealthChop
        self.set_spreadcycle(0)

        self._target_pos = self._steps_per_rev * revolutions
        self._step_interval = 0
        self._speed = 0.0
        self._n = 0
        self.set_acceleration(10000)
        self.set_max_speed(self._max_speed_homing)
        coolstep_thres = tmc_math.steps_to_tstep(self._max_speed_homing*0.5,
                                                 self.get_microstepping_resolution())
        self.set_coolstep_threshold(coolstep_thres)
        self.compute_new_speed()


        step_counter=0
        #self.tmc_logger.log("Steps per Revolution: {self._steps_per_rev}"")
        while step_counter<self._target_pos:
            if self.run_speed(): #returns true, when a step is made
                step_counter += 1
                self.compute_new_speed()
                sg_result = self.get_stallguard_result()
                sg_results.append(sg_result)
                if len(sg_results)>20:
                    sg_result_average = statistics.mean(sg_results[-6:])
                    if sg_result_average < self._sg_threshold:
                        break

        if step_counter<self._steps_per_rev:
            self.tmc_logger.log("homing successful",Loglevel.INFO)
            self.tmc_logger.log(f"Stepcounter: {step_counter}",Loglevel.DEBUG)
            self.tmc_logger.log(str(sg_results),Loglevel.DEBUG)
            self._current_pos = 0
        else:
            self.tmc_logger.log("homing failed", Loglevel.INFO)
            self.tmc_logger.log(f"Stepcounter: {step_counter}", Loglevel.DEBUG)
            self.tmc_logger.log(str(sg_results),Loglevel.DEBUG)

        self.tmc_logger.log("---", Loglevel.INFO)



    def reverse_direction_pin(self):
        """
        reverses the motor shaft direction
        """
        self._direction = not self._direction
        GPIO.output(self._pin_dir, self._direction)



    def set_direction_pin(self, direction):
        """
        sets the motor shaft direction to the given value: 0 = CCW; 1 = CW

            Parameters:
                direction (bool): motor shaft direction: 0 = CCW; 1 = CW
        """
        self._direction = direction
        GPIO.output(self._pin_dir, direction)



    def read_steps_per_rev(self):
        """
        returns how many steps are needed for one revolution.
        this reads the value from the tmc driver.

            Returns:
                steps_per_rev (int): Steps per revolution
        """
        self._steps_per_rev = self._fullsteps_per_rev*self.read_microstepping_resolution()
        return self._steps_per_rev



    def get_steps_per_rev(self):
        """
        returns how many steps are needed for one revolution.
        this gets the cached value from the library.

            Returns:
                steps_per_rev (int): Steps per revolution
        """
        return self._steps_per_rev



    def set_vactual_dur(self, vactual, duration=0, acceleration=0,
                             show_stallguard_result=False, show_tstep=False):
        """
        sets the register bit "VACTUAL" to to a given value
        VACTUAL allows moving the motor by UART control.
        It gives the motor velocity in +-(2^23)-1 [μsteps / t]
        0: Normal operation. Driver reacts to STEP input

            Parameters:
                vactual (int): value for vactual
                duration (int): after this vactual will be set to 0
                acceleration (int): use this for a velocity ramp
                show_stallguard_result (bool): prints StallGuard Result during movement
                show_tstep (bool): prints TStep during movement

            Returns:
                stop (enum): how the movement was finished
        """
        self._stop = StopMode.NO
        current_vactual = 0
        sleeptime = 0.05
        if vactual<0:
            acceleration = -acceleration

        if duration != 0:
            self.tmc_logger.log(f"vactual: {vactual} for {duration} sec",
                                Loglevel.INFO)
        else:
            self.tmc_logger.log(f"vactual: {vactual}", Loglevel.INFO)
        self.tmc_logger.log(str(bin(vactual)), Loglevel.INFO)

        self.tmc_logger.log("writing vactual", Loglevel.INFO)
        if acceleration == 0:
            self.set_vactual(int(round(vactual)))

        if duration == 0:
            return -1

        self._starttime = time.time()
        current_time = time.time()
        while current_time < self._starttime+duration:
            if self._stop == StopMode.HARDSTOP:
                break
            if acceleration != 0:
                time_to_stop = self._starttime+duration-abs(current_vactual/acceleration)
                if self._stop == StopMode.SOFTSTOP:
                    time_to_stop = current_time-1
            if acceleration != 0 and current_time > time_to_stop:
                current_vactual -= acceleration*sleeptime
                self.set_vactual(int(round(current_vactual)))
                time.sleep(sleeptime)
            elif acceleration != 0 and abs(current_vactual)<abs(vactual):
                current_vactual += acceleration*sleeptime
                self.set_vactual(int(round(current_vactual)))
                time.sleep(sleeptime)
            if show_stallguard_result:
                self.tmc_logger.log(f"StallGuard result: {self.get_stallguard_result()}",
                                    Loglevel.INFO)
                time.sleep(0.1)
            if show_tstep:
                self.tmc_logger.log(f"TStep result: {self.get_tstep()}",
                                    Loglevel.INFO)
                time.sleep(0.1)
            current_time = time.time()
        self.set_vactual(0)
        return self._stop



    def set_vactual_rps(self, rps, duration=0, revolutions=0, acceleration=0):
        """
        converts the rps parameter to a vactual value which represents
        rotation speed in revolutions per second
        With internal oscillator:
        VACTUAL[2209] = v[Hz] / 0.715Hz

            Parameters:
                rps (int): value for vactual in rps
                duration (int): after this vactual will be set to 0
                revolutions (int): after this vactual will be set to 0
                acceleration (int): use this for a velocity ramp

            Returns:
                stop (enum): how the movement was finished
        """
        vactual = tmc_math.rps_to_vactual(rps, self._steps_per_rev)
        if revolutions!=0:
            duration = abs(revolutions/rps)
        if revolutions<0:
            vactual = -vactual
        return self.set_vactual_dur(vactual, duration, acceleration=acceleration)



    def set_vactual_rpm(self, rpm, duration=0, revolutions=0, acceleration=0):
        """
        converts the rps parameter to a vactual value which represents
        rotation speed in revolutions per minute

            Parameters:
                rpm (int): value for vactual in rpm
                duration (int): after this vactual will be set to 0
                revolutions (int): after this vactual will be set to 0
                acceleration (int): use this for a velocity ramp

            Returns:
                stop (enum): how the movement was finished
        """
        return self.set_vactual_rps(rpm/60, duration, revolutions, acceleration)



    def set_stallguard_callback(self, pin_stallguard, threshold, callback,
                                min_speed = 100, ignore_delay = 0):
        """
        set a function to call back, when the driver detects a stall
        via stallguard
        high value on the diag pin can also mean a driver error

            Parameters:
                pin_stallguard (int): pin needs to be connected to DIAG
                threshold (int): value for SGTHRS
                callback (function): will be called on StallGuard trigger
                min_speed (int): min speed [steps/s] for StallGuard
        """
        self.tmc_logger.log(f"setup stallguard callback on GPIO {pin_stallguard}",
                            Loglevel.INFO)
        self.tmc_logger.log(f"""StallGuard Threshold: {threshold}
                            minimum Speed: {min_speed}""", Loglevel.INFO)

        self.set_stallguard_threshold(threshold)
        self.set_coolstep_threshold(tmc_math.steps_to_tstep(
            min_speed, self.get_microstepping_resolution()))
        self._sg_delay = ignore_delay
        self._sg_callback = callback
        self._pin_stallguard = pin_stallguard

        GPIO.setup(self._pin_stallguard, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)

        GPIO.add_event_detect(self._pin_stallguard, GPIO.RISING, callback=self.stallguard_callback,
                              bouncetime=300)



    def stallguard_callback(self, gpio_pin):
        """
        the callback function for StallGuard.
        only checks whether the duration of the current movement is longer than
        _sg_delay and then calls the actual callback

            Parameters:
                gpio_pin (int): pin number of the interrupt pin
        """
        del gpio_pin
        if self._sg_callback is None:
            self.tmc_logger.log("StallGuard callback is None", Loglevel.DEBUG)
            return
        if time.time()<=self._starttime+self._sg_delay and self._sg_delay != 0:
            return
        self._sg_callback()



#pylint: disable=invalid-name
"""
 TMC2209_Raspberry_Pi math library
"""



def rps_to_vactual(rps, steps_per_rev, fclk = 12000000):
    """
    converts rps -> vactual

        Parameters:
            rps (float): revolutions per second
            steps_per_rev (int): steps per revolution
            fclk (int): clock speed

        Returns:
            vactual (int): value for vactual
    """
    return int(round(rps / (fclk / 16777216) * steps_per_rev))


def vactual_to_rps(vactual, steps_per_rev, fclk = 12000000):
    """
    converts vactual -> rps

        Parameters:
            vactual (int): value for vactual
            steps_per_rev (int): steps per revolution
            fclk (int): clock speed

        Returns:
            rps (float): revolutions per second
    """
    return vactual * (fclk / 16777216) / steps_per_rev


def rps_to_steps(rps, steps_per_rev):
    """
    converts rps -> steps/second

        Parameters:
            rps (float): revolutions per second
            steps_per_rev (int): steps per revolution

        Returns:
            steps (int): steps per second
    """
    return rps * steps_per_rev

def steps_to_rps(steps, steps_per_rev):
    """
    converts steps/second -> rps

        Parameters:
            steps (int): steps per second
            steps_per_rev (int): steps per revolution

        Returns:
            rps (float): revolutions per second
    """
    return steps / steps_per_rev


def rps_to_tstep(rps, steps_per_rev, msres):
    """
    converts rps -> tstep

        Parameters:
            rps (float): revolutions per second
            steps_per_rev (int): steps per revolution
            msres (int): microstep resolution

        Returns:
            tstep (int): time per step
    """
    return int(round(12000000 / (rps_to_steps(rps, steps_per_rev) * 256 / msres)))


def steps_to_tstep(steps, msres):
    """
    converts steps/second -> tstep

        Parameters:
            steps (int): steps per second
            msres (int): microstep resolution

        Returns:
            tstep (int): time per step
    """
    return int(round(12000000 / (steps * 256 / msres)))



#pylint: disable=invalid-name
#pylint: disable=missing-function-docstring
#pylint: disable=missing-class-docstring
#pylint: disable=wildcard-import
#pylint: disable=pointless-statement
#pylint: disable=unused-wildcard-import
#pylint: disable=bare-except
"""
test for TMC_2209_move.py
"""

import unittest
from src.TMC_2209.TMC_2209_StepperDriver import *

class TestTMCMove(unittest.TestCase):

    def setUp(self):
        self.tmc = TMC_2209(21, 16, 20, serialport=None, skip_uart_init=True)
        self.tmc.set_acceleration_fullstep(100000)
        self.tmc.set_max_speed_fullstep(10000)
        self.tmc.set_movement_abs_rel(MovementAbsRel.ABSOLUTE)

    def tearDown(self):
        self.tmc.set_deinitialize_true

    def test_run_to_position_steps(self):

        self.tmc.run_to_position_steps(400, MovementAbsRel.RELATIVE)
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 400, "run_to_position_steps relative movement is wrong")

        self.tmc.run_to_position_steps(-200, MovementAbsRel.RELATIVE)
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 200, "run_to_position_steps relative movement is wrong")

        self.tmc.run_to_position_steps(400)
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 400, "run_to_position_steps absolute movement is wrong")

    def test_run_to_position_steps_threaded(self):
        self.tmc.run_to_position_steps_threaded(400, MovementAbsRel.RELATIVE)
        self.tmc.wait_for_movement_finished_threaded()
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 400, "run_to_position_steps_threaded relative movement is wrong")

        self.tmc.run_to_position_steps_threaded(-200, MovementAbsRel.RELATIVE)
        self.tmc.wait_for_movement_finished_threaded()
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 200, "run_to_position_steps_threaded relative movement is wrong")

        self.tmc.run_to_position_steps_threaded(400)
        self.tmc.wait_for_movement_finished_threaded()
        pos = self.tmc.get_current_position()
        self.assertEqual(pos, 400, "run_to_position_steps_threaded absolute movement is wrong")

        self.tmc.run_to_position_steps_threaded(800)
        time.sleep(0.05)
        self.tmc.stop()
        pos = self.tmc.get_current_position()
        print(f"motorposition: {pos}")
        self.assertTrue(400 < pos < 800, "run_to_position_steps_threaded stop movement is wrong")

if __name__ == '__main__':
    unittest.main()
